type tvar = int
type texp = 
    Tint
  | Tvar of tvar
  | Tarrow of texp * texp;;

type subst = (tvar * texp) list;;

let rec libre x = function 
    Tint -> false
  | Tvar n -> x = n
  | Tarrow (t1, t2) -> libre x t1 || libre x t2;;

let rec subst_tvar s x =  try List.assoc x s with Not_found -> Tvar x
let rec subst_texp s = function 
    Tint -> Tint
  | Tvar x -> subst_tvar s x
  | Tarrow (t1, t2) -> Tarrow (subst_texp s t1, subst_texp s t2);;

let subst_tenv s env =
  List.map (function x, t -> (x, subst_texp s t)) env
;;

let domain s = List.map (function x, t -> x) s;;
let restrict_out s l =
  List.filter (function x, t -> not (List.mem x l)) s;;
let rec compose s2 s1 =
  let s12 = List.map (fun (x, t) -> (x, subst_texp s1 t)) s2 in
  s12 @ restrict_out s2 (domain s1)
;;


exception Unify of texp * texp;;
let rec unify_list l =
  match l with
    [] -> []
  | (t1,t2)::l ->
      let s = unify_list l in 
      let s' = unify (subst_texp s t1) (subst_texp s t2) in
      compose s' s

and unify t1 t2 =
  match t1, t2 with
  | Tvar x, Tvar y when x = y -> []
  | Tvar x, t ->
      if libre x t then raise (Unify (t1, t2)) else [ x, t ]
  | t, Tvar x -> 
      if libre x t then raise (Unify (t1, t2)) else [ x, t ]
  | Tint, Tint -> [ ]
  | Tarrow (t11, t12), Tarrow (t21, t22) ->
      unify_list [ t11, t21; t12, t22 ]
  | t1, t2 -> raise (Unify (t1,t2))
;;

unify (Tarrow (Tint, Tvar 1)) (Tarrow (Tint, Tvar 2));;
